﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace RCM {
    public partial class ConfigForm : Form {
        #region Form
        protected override CreateParams CreateParams { // Workaround for control flickering reduction
            get {
                CreateParams cp = base.CreateParams;
                cp.ExStyle |= 0x02000000;  // WS_EX_COMPOSITED
                return cp;
            }
        }

        public ConfigForm() {
            InitializeComponent();
            treeView.ImageList = Config.IconList;
            treeView.ContextMenuStrip = GetNewDataSourceContextMenu();
            foreach (IDataSource dataSource in Config.DataSources) {
                string dataSourceType = dataSource.GetType().FullName;
                treeView.Nodes.Add(CreateTreeNodeFromDataSource(dataSource));
                SortTreeAndSelectNode(null);
            }
            foreach (IConfigSection configSection in Config.ConfigSections.Values) {
                TabPage tabPage = new TabPage(configSection.GetType().GetDisplayName());
                tabPage.Controls.Add(new ConfigFormTab(configSection));
                tabControl.TabPages.Add(tabPage);
            }
        }
        #endregion

        #region TreeView display
        private TreeNode CreateTreeNodeFromDataSource(IDataSource dataSource) {
            string imageKey = dataSource.GetType().FullName;
            return new TreeNode(dataSource.Title) { Tag = dataSource, ImageKey = imageKey, SelectedImageKey = imageKey, ContextMenuStrip = GetContextMenu(dataSource) };
        }

        private void SortTreeAndSelectNode(TreeNode node) {
            treeView.Sort();
            if (treeView.SelectedNode == node) {
                ChangeSelectedDataSource();
            } else {
                treeView.SelectedNode = node;
            }
        }
        #endregion

        #region TreeView navigation & selection
        private void TreeView_AfterSelect(object sender, TreeViewEventArgs e) => ChangeSelectedDataSource();

        private void TreeView_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e) => treeView.SelectedNode = e.Node;

        private void TreeView_MouseUp(object sender, MouseEventArgs e) {
            if (treeView.HitTest(e.X, e.Y).Node == null) {
                treeView.SelectedNode = null;
                ChangeSelectedDataSource();
            }
        }

        private void TreeView_KeyDown(object sender, KeyEventArgs e) {
            if (treeView.SelectedNode == null) {
                return;
            }
            if (e.KeyCode == Keys.F2) {
                treeView.SelectedNode.BeginEdit();
            } else if (e.KeyCode == Keys.Delete) {
                DeleteSelectedDataSource();
            }
        }

        private void ChangeSelectedDataSource() {
            bool isSelected = treeView.SelectedNode != null;
            textBoxTitle.Enabled = isSelected;
            textBoxTitle.Text = "";
            buttonSave.Enabled = isSelected;
            LoadDataSourceControl();
        }

        private void LoadDataSourceControl() {
            foreach (Control control in panelDataSource.Controls) {
                panelDataSource.Controls.Remove(control);
                control.Dispose();
            }
            if (treeView.SelectedNode != null) {
                IDataSource dataSource = (IDataSource)treeView.SelectedNode.Tag;
                textBoxTitle.Text = dataSource.Title;
                panelDataSource.Controls.Add(dataSource.GetControl());
            }
        }
        #endregion

        #region DataSource create/update/delete
        public delegate void AfterLabelEditWorkaround(TreeNode node);

        private void TreeView_AfterLabelEdit(object sender, NodeLabelEditEventArgs e) {
            if (e.Label == null) {
                e.CancelEdit = true;
                return;
            }
            IDataSource dataSource = (IDataSource)e.Node.Tag;
            dataSource.Title = e.Label;
            Config.Save();
            BeginInvoke(new AfterLabelEditWorkaround(SortTreeAndSelectNode), new object[] { e.Node }); // Workaround for treeView.Sort() triggering relabel on every sorted node. See note in https://docs.microsoft.com/en-us/dotnet/api/system.windows.forms.listview.afterlabeledit?view=netframework-4.6
        }

        private void ButtonSave_Click(object sender, EventArgs e) {
            TreeNode selectedNode = treeView.SelectedNode;
            IDataSource dataSource = (IDataSource)selectedNode.Tag;
            dataSource.UpdateFromControl();
            dataSource.Title = selectedNode.Text = textBoxTitle.Text;
            Config.Save();
            SortTreeAndSelectNode(selectedNode);
        }

        private void NewDataSourceContextMenuItem_Click(object sender, EventArgs e) {
            Type type = (Type)((ToolStripMenuItem)sender).Tag;
            IDataSource dataSource = (IDataSource)Activator.CreateInstance(type);
            dataSource.Title = string.Format("New {0}", type.GetDisplayName());
            InsertNewDataSource(dataSource);
        }

        private void DuplicateDataSourceContextMenuItem_Click(object sender, EventArgs e) {
            IDataSource source = (IDataSource)treeView.SelectedNode.Tag;
            IDataSource dataSource = (IDataSource)ConfigSerializer.JsonToObj(ConfigSerializer.ObjToJson(source));
            dataSource.Title = string.Format("{0} - Copy", dataSource.Title);
            InsertNewDataSource(dataSource);
        }

        private void InsertNewDataSource(IDataSource dataSource) {
            Config.DataSources.Add(dataSource);
            Config.Save();
            TreeNode node = CreateTreeNodeFromDataSource(dataSource);
            treeView.Nodes.Add(node);
            SortTreeAndSelectNode(node);
        }

        private void DeleteSelectedDataSource() {
            IDataSource dataSource = (IDataSource)treeView.SelectedNode.Tag;
            if (MessageBox.Show("Do you really want to delete selected data source?", "Confirmation", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes) {
                Config.DataSources.Remove(dataSource);
                Config.Save();
                treeView.SelectedNode.Remove();
                SortTreeAndSelectNode(null);
            }
        }

        private void DeleteDataSourceContextMenuItem_Click(object sender, EventArgs e) => DeleteSelectedDataSource();
        #endregion

        #region DataSource ContextMenu
        private ContextMenuStrip GetNewDataSourceContextMenu() {
            ContextMenuStrip menu = new ContextMenuStrip();
            ToolStripMenuItem menuItem = new ToolStripMenuItem("New item");
            foreach (Type type in Config.DataSourceTypes) {
                ToolStripMenuItem item = new ToolStripMenuItem(type.GetDisplayName()) { Tag = type };
                item.Click += NewDataSourceContextMenuItem_Click;
                menuItem.DropDownItems.Add(item);
            }
            menu.Items.Add(menuItem);
            return menu;
        }

        private ContextMenuStrip GetContextMenu(IDataSource dataSource) {
            ContextMenuStrip menu = new ContextMenuStrip();
            ToolStripMenuItem duplicateItem = new ToolStripMenuItem("Duplicate item");
            duplicateItem.Click += DuplicateDataSourceContextMenuItem_Click;
            menu.Items.Add(duplicateItem);
            ToolStripMenuItem deleteItem = new ToolStripMenuItem("Delete item");
            deleteItem.Click += DeleteDataSourceContextMenuItem_Click;
            menu.Items.Add(deleteItem);
            return menu;
        }
        #endregion
    }
}
